/**
 * \file control/cards.c
 * \brief Basic Soundcard Operations
 * \author Jaroslav Kysela <perex@perex.cz>
 * \date 1998-2001
 */
/*
 *  Soundcard Operations - main file
 *  Copyright (c) 1998 by Jaroslav Kysela <perex@perex.cz>
 *
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as
 *   published by the Free Software Foundation; either version 2.1 of
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include "control_local.h"

#ifndef DOC_HIDDEN

#define SND_FILE_CONTROL_FMT		ALSA_DEVICE_DIRECTORY "controlC%i"
#define SIZE_OF_SND_FILE_CONTROL_STR	( sizeof( SND_FILE_CONTROL_FMT ) + 10 )

#define SND_FILE_ALOAD_FMT		ALOAD_DEVICE_DIRECTORY "aloadC%i"
#define SIZE_OF_SND_FILE_ALOAD_STR	( sizeof( SND_FILE_ALOAD_FMT ) + 10 )

#endif /* DOC_HIDDEN */


static inline char * snd_card_print_path( char *buffer, const char *control_path_fmt, int card ){
	sprintf( buffer, control_path_fmt, card );
	return buffer;
}

#ifndef DOC_HIDDEN

/* defining context-dependent calls for snd_card_print_path */
#define _snd_card_build_control_path(buff, card)		snd_card_print_path( (buff), SND_FILE_CONTROL_FMT, (card) )
#define _snd_card_build_aload_path(buff, card)		snd_card_print_path( (buff), SND_FILE_ALOAD_FMT, (card) )

#define DONT_PRINT_ERRORS	0
#define PRINT_ERRORS	1

#endif /* DOC_HIDDEN */

static int snd_card_load2( const char *control, snd_ctl_card_info_t * out_info, int prn_err )
{
	int open_dev;
	snd_ctl_card_info_t *info_ref;

	open_dev = snd_open_device(control, O_RDONLY);

	if (open_dev >= 0) {
		if( out_info ) /* a reference to the object returned by next ioctl has been asked by the caller */
			info_ref = out_info;
		else /* next ioctl is only relevant for internal implementation */
			info_ref = (snd_ctl_card_info_t *) alloca( sizeof( snd_ctl_card_info_t ) );
		if (ioctl(open_dev, SNDRV_CTL_IOCTL_CARD_INFO, info_ref) < 0) {
			int err = -errno;
			close(open_dev);
			if( out_info && prn_err ) /* if the caller required info, it might want info on errors too - behaving same as snd_ctl_hw_card_info */
				SYSERR("SNDRV_CTL_IOCTL_CARD_INFO failed");
			return err;
		}
		close(open_dev);
		return info_ref->card;
	} else {
		return -errno;
	}
}


static int snd_card_load1( int card, snd_ctl_card_info_t *info_ref, int prn_err )
{
	int res;
	char control_path[ SIZE_OF_SND_FILE_CONTROL_STR ];

	if( prn_err ){ /* behaving same as snd_ctl_hw_open on request */
		if (CHECK_SANITY(card < 0 || card >= 32)) {
			SNDMSG("Invalid card index %d", card);
			return -EINVAL;
		}
	}

	res = snd_card_load2( _snd_card_build_control_path( control_path, card ), info_ref, prn_err );

#ifdef SUPPORT_ALOAD
	if (res < 0) {
		char aload_path[ SIZE_OF_SND_FILE_ALOAD_STR ];
		res = snd_card_load2( _snd_card_build_aload_path( aload_path, card ), info_ref, prn_err );
	}
#endif /* SUPPORT_ALOAD */

	return res;
}


#ifndef	DOC_HIDDEN

/* main goal is validating card number -> call snd_card_load1 with appropriate name */
#define snd_card_load_value( card, info )	snd_card_load1( (card), (info), DONT_PRINT_ERRORS )

/* main goal is getting card info -> call snd_card_load1 with appropriate name */
#define snd_card_load_info( card, info )	snd_card_load1( (card), (info), PRINT_ERRORS )

/* a string value must be passed instead of a card number -> call snd_card_load2 appropriately */
#define snd_card_load_info_from_string( str, info )	snd_card_load2( (str), (info), PRINT_ERRORS )

#endif /* DOC_HIDDEN */


/**
 * \brief Try to load the driver for a card.
 * \param card Card number.
 * \return 1 if driver is present, zero if driver is not present
 */
int snd_card_load(int card)
{
	return !!( snd_card_load_value( card, NULL ) >= 0 );
}


/**
 * \brief Try to determine the next card.
 * \param rcard pointer to card number
 * \result zero if success, otherwise a negative error code
 *
 * Tries to determine the next card from given card number.
 * If card number is -1, then the first available card is
 * returned. If the result card number is -1, no more cards
 * are available.
 */
int snd_card_next(int *rcard)
{
	int card;
	
	if (rcard == NULL)
		return -EINVAL;
	card = *rcard;
	card = card < 0 ? 0 : card + 1;
	for (; card < 32; card++) {
		if (snd_card_load(card)) {
			*rcard = card;
			return 0;
		}
	}
	*rcard = -1;
	return 0;
}

/**
 * \brief Convert card string to an integer value.
 * \param string String containing card identifier
 * \return an integer value greater than or equal to zero, corresponding to the string identifier, if success, otherwise a negative error code.
 *
 * The accepted format is an integer value in ASCII representation
 * or the card identifier (the id parameter for sound-card drivers).
 * The control device name like /dev/snd/controlC0 is accepted, too.
 */
int snd_card_get_index(const char *string)
{
	int card, err;
	snd_ctl_card_info_t info;

	if (!string || *string == '\0')
		return -EINVAL;

	/* integer value */
	if ((isdigit(*string) && *(string + 1) == 0) ||
	    (isdigit(*string) && isdigit(*(string + 1)) && *(string + 2) == 0)) {
		if (sscanf(string, "%i", &card) != 1)
			return -EINVAL;
		if (card < 0 || card > 31)
			return -EINVAL;
		err = snd_card_load_value( card, NULL );
		if (err >= 0) /* can this be different from card? */
			return card;
		return err;
	}

	/* device name */
	if (string[0] == '/')	
		return snd_card_load_info_from_string( string, &info );

	/* card identifier */
	for (card = 0; card < 32; card++) {
		if( snd_card_load_info( card, &info ) < 0 )
			continue;
		if (!strcmp((const char *)info.id, string))
			return card;
	}
	return -ENODEV;
}


/**
 * \brief Obtain the card name.
 * \param card Card number
 * \param name Result - card name corresponding to card number
 * \result zero if success, otherwise a negative error code
 */
int snd_card_get_name(int card, char **name)
{
	snd_ctl_card_info_t info;
	int err;
	
	if (name == NULL)
		return -EINVAL;

	if( ( err = snd_card_load_info( card, &info ) ) < 0 )
		return err;

	*name = strdup((const char *)info.name);
	if (*name == NULL)
		return -ENOMEM;

	return 0;
}


/**
 * \brief Obtain the card long name.
 * \param card Card number
 * \param name Result - card long name corresponding to card number
 * \result zero if success, otherwise a negative error code
 */
int snd_card_get_longname(int card, char **name)
{
	snd_ctl_card_info_t info;
	int err;
	
	if (name == NULL)
		return -EINVAL;

	if( ( err = snd_card_load_info( card, &info ) ) < 0 )
		return err;

	*name = strdup((const char *)info.longname);
	if (*name == NULL)
		return -ENOMEM;

	return 0;
}